/*
 * tmos - sched.c
 * 	version 0.0.0
 *
 * 	Task scheduler
 * 
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/sched.h>
#include <tmos/kernel.h>

struct context task[CONFIG_TASKS_MAX];
struct context __kernel *next_context;

__bit __timers_pending = 0;

/*
 * timers
 */
void do_timers(void) __using(SOFTIRQ_BANK)
{
	struct context __kernel *tsk;

	/*
	 * execute task timers
	 */
	tsk = task;
	do {
		if (tsk->state >= TASK_SLEEP) {
			tsk->state++;
			if (task_is_running(tsk)) {
				local_irq_disable();
				__wake_up_context(tsk);
				local_irq_enable();
			}
		}
		tsk++;
	} while (tsk != task + CONFIG_TASKS_MAX);
}

/**
 * schedule - schedule task
 *
 * Currently only Round-Robin scheduling is implemented.
 *
 * This function should be called from kernel context with interrupts
 * enabled.
 */
void schedule(void)
{
	for (;;) {
		watchdog_schedule();
		local_irq_disable();
		if (current->next == current) {
			if (task_is_running(current)) {
				local_irq_enable();
				return;
			}

			/*
			 * current is the only task in runqueue and it's
			 * not running, so we can wait for interrupt
			 */
			safe_halt();
		} else {
			thread_switch();
			BUG_ON(!task_is_running(current));
			return;
		}
	}
}

/**
 * wake_up_context - add task to runqueue
 * @context: task to add
 *
 * This function should be called in kernel mode.
 */
void wake_up_context(struct context __kernel *context) __reentrant
{
	cpu_flags_t flags;

	local_irq_save(flags);
	__wake_up_context(context);
	local_irq_restore(flags);
}

/*
 * __remove_run_queue - remove context from run queue
 * @context: context to remove
 *
 * This function can be called only with interrupts disabled.
 */
void __remove_run_queue(struct context __kernel *context) __reentrant
{
	struct context __kernel *c;

	c = current;
	do {
		if (c->next == context) {
			c->next = context->next;
			return;
		}
		c = c->next;
	} while (c != current);
}

void sched_init(void)
{
	unsigned char i;

#ifdef CONFIG_TDATA
	for (i = 0; i < CONFIG_TASKS_MAX; i++)
		BUG_ON(page_alloc() != i + TDATA_OFFSET);
#endif
	for (i = 0; i < CONFIG_TASKS_MAX; i++)
		set_task_state(&task[i], TASK_DEAD);

	context_init(&task[0]);
	task[0].next = &task[0];
	current = &task[0];
	set_current_state(TASK_RUNNING);
}
